/*====================================+=====================================+
! File CServerList.cpp                ! Copyright (C) 2018-2019 Remi PASCAL !
+-------------------------------------+-------------------------------------+
! This file is part of kimfo.                                               !
! It is free software: you can redistribute it and/or modify it under the   !
! terms of the GNU General Public License as published by the Free Software !
! Foundation, either version 3 of the License, or any later version.        !
! This application is distributed in the hope that it will be useful, but   !
! WITHOUT ANY WARRANTY; without even the implied warranty of                !
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General !
! Public License for more details.                                          !
! You should have received a copy of the GNU General Public License along   !
! with this application. If not, see <http://www.gnu.org/licenses/>.        !
+==========================================================================*/



/*-------------------------------------------------------------------------*/
#include <wx/wx.h>
#include <wx/regex.h>
#include <wx/imaglist.h>
#include <wx/renderer.h>
#include <wx/tokenzr.h>
#include <wx/timer.h>
#include "CApp.h"
#include "CFrame.h"
#include "common/sc_download.h"
#include "CServerList.h"
#include "CDlgPrefs.h"
#include "check_16x16.xpm"
#include "uncheck_16x16.xpm"
#include "flags/canada_16x16.xpm"
#include "flags/france_16x16.xpm"
/*-------------------------------------------------------------------------*/



/*-------------------------------------------------------------------------*/
CServerList::CServerList( wxWindow *parent )
           : wxListCtrl( parent, wxID_ANY, wxDefaultPosition,
                         wxDefaultSize, wxLC_REPORT | wxLC_VIRTUAL
                       )
{
   /*--( Regions alphabetically sorted on the region name )----------------*/
      m_vec_region
   =  {
       { "ca", "Canada [EN]", "ca/en/servers.xml",
         "ca/en/order/kimsufi.xml?reference="
       },
       { "qc", "Canada [FR]", "ca/fr/serveurs.xml",
         "ca/fr/commande/kimsufi.xml?reference="
       },
       { "cz", wxString::FromUTF8( u8"\u010Cesk\u00E1 Republika" ),
         "cz/servery.xml", "cz/objednavka/kimsufi.xml?reference="
       },
       { "de", "Deutschland",
         "de/server.xml", "de/bestellung/isgenug.xml?reference="
       },
       { "es", wxString::FromUTF8( u8"Espa\u00F1a" ),
         "es/servidores.xml", "es/pedido/kemsirve.xml?reference="
       },
       { "fr", "France",
         "fr/serveurs.xml", "fr/commande/kimsufi.xml?reference="
       },
       { "ie", "Ireland",
         "en/servers.xml", "en/order/kimsufi.xml?reference="
       },
       { "it", "Italia",
         "it/server.xml", "it/ordine/kimsufi.xml?reference="
       },
       { "lt", "Lietuva",
         "lt/serveriai.xml", "lt/uzsakymas/kimsufi.xml?reference="
       },
       { "ma", "Maroc",
         "ma/serveurs.xml", "ma/commande/kimsufi.xml?reference="
       },
       { "nl", "Nederland",
         "nl/servers.xml", "nl/bestelling/isgenoeg.xml?reference="
       },
       { "pl", "Polska",
         "pl/serwery.xml", "pl/zamowienie/kimsufi.xml?reference="
       },
       { "pt", "Portugal",
         "pt/servidores.xml", "pt/encomenda/kemeda.xml?reference="
       },
       { "sn", wxString::FromUTF8( u8"S\u00e9n\u00e9gal" ),
         "sn/serveurs.xml",
       },
       { "fi", "Suomi",
         "fi/palvelimet.xml", "fi/tilaus/kimsufi.xml?reference="
       },
       { "tn", "Tunisie",
         "tn/serveurs.xml", "tn/commande/kimsufi.xml?reference="
       },
       { "uk", "United Kingdom",
         "uk/servers.xml", "uk/order/kimsufi.xml?reference="
       },
       /*--( World in $ )--------------------------------------------------*/
       { "W$", "World [$]",
         "us/en/servers.xml", "us/en/order/kimsufi.xml?reference="
       },
       /*--( World in Euros )----------------------------------------------*/
       { "WE", wxString::FromUTF8( u8"World [\u20AC]" ),
         "en/servers.xml", "en/order/kimsufi.xml?reference="
       }
     } ;

   /*----------------------------------------------------------------------*/
   AppendColumn( wxEmptyString, wxLIST_FORMAT_LEFT, 28 ) ; // Check box
   AppendColumn( _( "Name" ), wxLIST_FORMAT_LEFT, 60 ) ;
   AppendColumn( _( "Reference" ) ) ;
   AppendColumn( _( "Data center" ), wxLIST_FORMAT_LEFT, 120 ) ;
   AppendColumn( _( "Price" ), wxLIST_FORMAT_LEFT, 190 ) ;
   AppendColumn( _( "Specification" ), wxLIST_FORMAT_LEFT, 1000 ) ;

   /*----------------------------------------------------------------------*/
   string_to_col_width( wxGetApp().M_s_serv_list_col_width.get() ) ;
   /*----------------------------------------------------------------------*/
   m_il = new wxImageList( 16, 16 ) ;
   add_checkboxes() ;

   /*--( Country data center )---------------------------------------------*/
   m_map_dc_ico[ "bhs" ] = add_img_il( wxBitmap( canada_16x16_xpm ) ) ;
   m_map_dc_ico[ "gra rbx sbg" ] = add_img_il( wxBitmap( france_16x16_xpm ));

   /*----------------------------------------------------------------------*/
   SetImageList( m_il, wxIMAGE_LIST_SMALL ) ;

   /*----------------------------------------------------------------------*/
   m_timer_check_avail.SetOwner( this ) ;
   Bind( wxEVT_TIMER,
         [ this ]( wxTimerEvent &event ) { launch_check_avail() ; },
         m_timer_check_avail.GetId()
      ) ;
   /*----------------------------------------------------------------------*/
   Bind( wxEVT_LIST_ITEM_ACTIVATED,
         [ this ]( wxListEvent &event )
         {  /*-------------------------------------------------------------*/
            const wxString s_url = get_url( event.GetIndex() ) ;
            if( !s_url.empty() ) { wxLaunchDefaultBrowser( s_url ) ; }
            /*-------------------------------------------------------------*/
         }
       ) ;
   Bind( wxEVT_LEFT_DOWN, &CServerList::OnMouseLeftDown, this ) ;
   Bind( wxEVT_KEY_DOWN, &CServerList::OnKeyDown, this ) ;

   /*-----------------------------------------------------------------------+
   ! Download management                                                    !
   +-----------------------------------------------------------------------*/
   Bind( sc::EVT_DOWNLOAD,
         [ this ]( sc::CDownloadEvent &event )
         {  /*-------------------------------------------------------------*/
            switch( event.get_download_id() )
            {
               /*----------------------------------------------------------*/
               case DID_SERVER :
                  refresh_server( event.get_err_code(), event.get_string() );
                  break ;
               /*----------------------------------------------------------*/
               case DID_AVAIL :
                  check_avail( event.get_err_code(),
                               event.get_user_data(),
                               event.get_string()
                             ) ;
                  break ;
               /*----------------------------------------------------------*/
               default :
                  wxFAIL_MSG( "Unmanaged download id" ) ;
                  break ;
               /*----------------------------------------------------------*/
            }
            /*-------------------------------------------------------------*/
         }
         /*----------------------------------------------------------------*/
      ) ;

   /*--( Selected and available item )-------------------------------------*/
   const wxColour co_coul_ok( 115, 230, 0 ) ;
   m_a_lia[ 0 ][ 0 ] = wxItemAttr( wxNullColour, co_coul_ok, GetFont() ) ;
   m_a_lia[ 0 ][ 1 ] = wxItemAttr( wxNullColour,
                                   sc::alt_bg_color( co_coul_ok ), GetFont()
                                 ) ;
   /*--( Selected and NOT available item )---------------------------------*/
   const wxColour co_coul_ko( 255, 102, 102 ) ;
   m_a_lia[ 1 ][ 0 ] = wxItemAttr( wxNullColour, co_coul_ko, GetFont() ) ;
   m_a_lia[ 1 ][ 1 ] = wxItemAttr( wxNullColour,
                                   sc::alt_bg_color( co_coul_ko ), GetFont()
                                 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CServerList::~CServerList()
{
   /*----------------------------------------------------------------------*/
   delete m_il ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxItemAttr *CServerList::OnGetItemAttr( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   /*----------------------------------------------------------------------*/
   if( p_server == nullptr ) { return nullptr ; }
   /*----------------------------------------------------------------------*/
   if( !p_server->m_boo_select || p_server->m_avail == CServer::AVA_UNKNOWN )
   {  return nullptr ; }
   /*----------------------------------------------------------------------*/
   const int co_i_lia = ( p_server->m_avail == CServer::AVA_OK ? 0 : 1 ) ;
   return const_cast< wxItemAttr * >( &m_a_lia[ co_i_lia ][ l_item % 2 ] ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::OnGetItemColumnImage( long l_item, long l_column ) const
{
   /*----------------------------------------------------------------------*/
   const int co_i_no_image = -1 ;
   const CServer *p_server = get_server( l_item ) ;
   /*----------------------------------------------------------------------*/
   if( p_server == nullptr ) { return co_i_no_image ; }
   /*----------------------------------------------------------------------*/
   switch( l_column )
   {
      /*-------------------------------------------------------------------*/
      case COL_CHECK :
         return p_server->m_boo_select ? m_i_img_check : m_i_img_uncheck ;
      /*-------------------------------------------------------------------*/
      case COL_DC :
      {  auto it = m_map_dc_ico.find( p_server->m_s_dc ) ;
         return it != m_map_dc_ico.end() ? it->second : co_i_no_image ;
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return co_i_no_image ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::OnGetItemText( long l_item, long l_column ) const
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   /*----------------------------------------------------------------------*/
   if( p_server == nullptr ) { return wxEmptyString ; }
   /*----------------------------------------------------------------------*/
   switch( l_column )
   {
      /*-------------------------------------------------------------------*/
      case COL_NAME  : return p_server->m_s_name ;
      case COL_REF   : return p_server->m_s_ref ;
      case COL_DC    : return p_server->m_s_dc ;
      case COL_PRICE : return p_server->m_s_price ;
      case COL_SPEC  : return p_server->m_s_spec ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return wxEmptyString ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::OnMouseLeftDown( wxMouseEvent &event )
{
   /*----------------------------------------------------------------------*/
   int  flags ;
   const long co_l_item = HitTest( event.GetPosition(), flags ) ;

   /*--( Click in the checkbox ? )-----------------------------------------*/
   if( co_l_item > -1 && ( flags & wxLIST_HITTEST_ONITEMICON ) != 0 )
   {
      /*-------------------------------------------------------------------*/
      CServer *p_server = get_server( co_l_item ) ;
      if( p_server == nullptr ) { return ; }
      /*-------------------------------------------------------------------*/
      p_server->m_boo_select = !p_server->m_boo_select ;
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   event.Skip() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::OnKeyDown( wxKeyEvent &event )
{
   /*----------------------------------------------------------------------*/
   bool boo_cmd_key
      = ( ( event.GetModifiers() & ~wxMOD_ALT ) == wxMOD_CMD ) ;
   bool boo_alt_key
      = ( ( event.GetModifiers() & wxMOD_ALT ) == wxMOD_ALT ) ;
   /*----------------------------------------------------------------------*/
   if( !boo_alt_key && !boo_cmd_key && event.GetKeyCode() == ' ' )
   {  select_unselect_hil_following_focus() ; }
   else
   {  event.Skip() ; }
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! Note : The checkbox can be, at least under XFCE, larger than 16x16        !
+---------------------------------------------------------------------------+
! I have decided to stay to "16x16" image list size for different reasons : !
! - If the checkboxes are smaller we can use them centered in the 16x16     !
! - Bigger image size doesn't seem to be supported under MSW                !
! - Under XFCE, wx uses a weird mix of icon sizes : 16x16 seeming to be the !
!   one used for other icons than the first one (checkbox)                  !
! - I would have liked to use the "native checkbox feature" appeared in     !
!   wx 3.1.0 but it does'nt seem, for now, to work in a virtual file list   !
+---------------------------------------------------------------------------+
! Note : Rescaling checkboxes to 16x16 doesn't give nice results ...        !
+--------------------------------------------------------------------------*/
void CServerList::add_checkboxes()
{
   /*----------------------------------------------------------------------*/
   wxRendererNative &renderer = wxRendererNative::Get() ;
   const wxSize co_size_il = m_il->GetSize() ;
   const wxSize co_size_cb = renderer.GetCheckBoxSize( this ) ;

   /*--( If they are bigger than expected we don't use system checkboxes )-*/
   if(    co_size_cb.GetWidth()  > co_size_il.GetWidth()
       && co_size_cb.GetHeight() > co_size_il.GetHeight()
     )
   {
      /*-------------------------------------------------------------------*/
      m_i_img_check   = m_il->Add( wxIcon( check_16x16_xpm ) ) ;
      m_i_img_uncheck = m_il->Add( wxIcon( uncheck_16x16_xpm ) ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   {
      /*--( The "system" checkboxes have a "compatible" size )-------------*/
      wxBitmap bmp_cb( co_size_il ) ;
      /*--( If the box is smaller than planned, we have to center it )-----*/
      const wxPoint co_pt( (   co_size_il.GetWidth()
                             - co_size_cb.GetWidth()
                           ) / 2,
                           (   co_size_il.GetHeight()
                             - co_size_cb.GetHeight()
                           ) / 2
                         ) ;
      const wxRect co_rect( co_pt, co_size_cb ) ;
      wxMemoryDC   dc ;
      /*-------------------------------------------------------------------*/
      dc.SelectObject( bmp_cb ) ;
      dc.Clear() ;
      renderer.DrawCheckBox( this, dc, co_rect,
                             wxCONTROL_FLAT | wxCONTROL_CHECKED
                           ) ;
      dc.SelectObject( wxNullBitmap ) ;
      /*----------------------------------------------------------------*/
      m_i_img_check = m_il->Add( bmp_cb ) ;

      /*-------------------------------------------------------------------*/
      dc.SelectObject( bmp_cb ) ;
      dc.Clear() ;
      renderer.DrawCheckBox( this, dc, co_rect, wxCONTROL_FLAT ) ;
      dc.SelectObject( wxNullBitmap ) ;
      /*-------------------------------------------------------------------*/
      m_i_img_uncheck = m_il->Add( bmp_cb ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*--------------------------------------------------------------------------+
! This function will only create "fake" selected servers that automatically !
! will be reselected after first server refresh                             !
+--------------------------------------------------------------------------*/
void CServerList::string_to_serv_ref_sel( const wxString &s_val )
{
   /*----------------------------------------------------------------------*/
   wxStringTokenizer tkz( s_val, '\n' ) ;
   wxString s_token ;
   /*----------------------------------------------------------------------*/
   while( tkz.HasMoreTokens() )
   {
      /*-------------------------------------------------------------------*/
      m_vec_server.emplace_back( tkz.GetNextToken() ) ;
      m_vec_server.back().m_boo_select = true ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::refresh_all_items()
{
   /*----------------------------------------------------------------------*/
   const int co_i_nb = GetItemCount() ;
   /*----------------------------------------------------------------------*/
   if( co_i_nb <= 0 ) { return ; }
   /*----------------------------------------------------------------------*/
   RefreshItems( 0, co_i_nb - 1 ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::serv_ref_sel_to_string() const
{
   /*----------------------------------------------------------------------*/
   wxString s_res ;
   /*----------------------------------------------------------------------*/
   for( const auto &server : m_vec_server )
   {  /*-------------------------------------------------------------------*/
      if( !server.m_boo_select ) { continue ; }
      /*-------------------------------------------------------------------*/
      if( !s_res.empty() ) { s_res += '\n' ; }
      s_res += server.m_s_ref ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return s_res ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::set_select( long l_item, bool boo_select )
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= static_cast< int >( m_vec_server.size() ) )
   {  return -1 ; }
   /*----------------------------------------------------------------------*/
   m_vec_server[ l_item ].m_boo_select = boo_select ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::select_all( bool boo_select )
{
   /*----------------------------------------------------------------------*/
   for( CServer &server : m_vec_server )
   {  server.m_boo_select = boo_select ; }
   /*----------------------------------------------------------------------*/
   refresh_all_items() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::select_invert()
{
   /*----------------------------------------------------------------------*/
   for( CServer &server : m_vec_server )
   {  server.m_boo_select = !server.m_boo_select ; }
   /*----------------------------------------------------------------------*/
   refresh_all_items() ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::select_unselect_hil_following_focus()
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( get_item_focus() ) ;
   if( p_server == nullptr ) { return ; }
   /*----------------------------------------------------------------------*/
   const bool co_boo_select = !p_server->m_boo_select ;
   /*----------------------------------------------------------------------*/
   for( long l_item = -1 ; ( l_item = get_next_item_hil( l_item ) ) != -1 ; )
   {  set_select( l_item, co_boo_select ) ; RefreshItem( l_item ) ; }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
const CServer *CServerList::get_server( long l_item ) const
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= static_cast< int >( m_vec_server.size() ) )
   {  return nullptr ; }
   /*----------------------------------------------------------------------*/
   return &m_vec_server[ l_item ] ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
CServer *CServerList::get_server( long l_item )
{
   /*----------------------------------------------------------------------*/
   if( l_item < 0 || l_item >= static_cast< int >( m_vec_server.size() ) )
   {  return nullptr ; }
   /*----------------------------------------------------------------------*/
   return &m_vec_server[ l_item ] ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::add_img_il( const wxBitmap &bmp )
{
   /*----------------------------------------------------------------------*/
   const wxSize co_size_bmp( bmp.GetSize() ) ;
   const wxSize co_size_il( m_il->GetSize() ) ;

   /*--( Same sizes ? the image is added )---------------------------------*/
   if( co_size_bmp == co_size_il ) { return m_il->Add( bmp ) ; }

   /*--( If image is smaller it is centered in image list size space )-----*/
   if(    co_size_bmp.GetHeight() <= co_size_il.GetHeight()
       && co_size_bmp.GetWidth()  <= co_size_il.GetWidth()
     )
   {  /*--( Resize the image to the image list image expected size )-------*/
      wxPoint pt( ( co_size_il.GetWidth()  - co_size_bmp.GetWidth() ) / 2,
                  ( co_size_il.GetHeight() - co_size_bmp.GetHeight() ) / 2
                ) ;
      /*-------------------------------------------------------------------*/
      return m_il->Add( bmp.ConvertToImage().Resize( co_size_il, pt ) ) ;
      /*-------------------------------------------------------------------*/
   }

   /*--( Else the image is rescaled )--------------------------------------*/
   return m_il->Add( sc::rescale( bmp, co_size_il ) ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
void CServerList::string_to_col_width( const wxString &s_val )
{
   /*----------------------------------------------------------------------*/
   const int co_i_nb_col = GetColumnCount() ;
   wxStringTokenizer tkz( s_val, '\n' ) ;
   wxString s_token ;
   /*----------------------------------------------------------------------*/
   for( int i_col = 0 ; i_col < co_i_nb_col && tkz.HasMoreTokens() ; ++i_col)
   {
      /*-------------------------------------------------------------------*/
      int i_width = wxAtoi( tkz.GetNextToken() ) ;
      /*--( Minimal check )------------------------------------------------*/
      if( i_width >= 10 )
      {  SetColumnWidth( i_col, i_width ) ; }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::col_width_to_string() const
{
   /*----------------------------------------------------------------------*/
   wxString s_res ;
   const int co_i_nb_col = GetColumnCount() ;
   /*----------------------------------------------------------------------*/
   for( int i_col = 0 ; i_col < co_i_nb_col ; ++i_col )
   {
      /*-------------------------------------------------------------------*/
      if( !s_res.empty() ) { s_res += '\n' ; }
      /*-------------------------------------------------------------------*/
      s_res += wxString::Format( "%d", GetColumnWidth( i_col ) ) ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   return s_res ;
   /*----------------------------------------------------------------------*/
}

/*--( Starting the download of the server web page )-----------------------*/
int CServerList::launch_refresh_server()
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_title
       = wxString::Format( _( "Refresh server list #%d" ),
                           ++m_i_nb_refresh_server
                         ) ;
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   frame->gui_enable_refresh_server( false ) ;
   frame->gui_enable_check_avail( false ) ;
   /*----------------------------------------------------------------------*/
   const str_region *p_region = get_region() ;
   if( p_region == nullptr )
   {  frame->log_message( wxString::Format( _( "%s: region non found" ),
                                            co_s_title
                                          )
                        ) ;
   }
   /*----------------------------------------------------------------------*/
   frame->log_message( wxString::Format( _( "%s: region \"%s\""),
                                         co_s_title, p_region->m_s_name
                                       )
                     ) ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_url
                = wxString::Format( "%s/%s", m_co_p_c_main_url,
                                    p_region->m_p_c_url_server
                                  ) ;
   /*----------------------------------------------------------------------*/
   if( co_s_url.empty() || co_s_url.size() <= 10 )
   {  frame->log_message( wxString::Format( _( "%s: invalid url" ),
                                            co_s_title
                                          )
                        ) ;
   }
   /*----------------------------------------------------------------------*/
   sc::CDownload dl( this, co_s_url ) ;
   dl.get_dlbase().set_download_id( DID_SERVER ) ;
   dl.thread_start() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::refresh_server( int i_err_code, const wxString &co_s_data )
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_title = _( "Servers" ) ;
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   frame->gui_enable_refresh_server( true ) ;
   frame->gui_enable_check_avail( true ) ;
   /*----------------------------------------------------------------------*/
   if( i_err_code != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  frame->log_error( sc::CDownloadBase::st_get_err_msg( i_err_code ) ) ;
      return -1 ;
   }
   /*----------------------------------------------------------------------*/
   if( co_s_data.empty() )
   {  frame->log_message( wxString::Format( _( "%s: no data" ), co_s_title
                                          )
                        ) ;
   }
   /*--( Save and reset previous server data )-----------------------------*/
   t_vec_server vec_server_save ;
   m_vec_server.swap( vec_server_save ) ;
   /*--( Replace "unbreakable space" characters )--------------------------*/
   wxString s_data = co_s_data ;
   s_data.Replace( wxString::FromUTF8( u8"\u00A0" ), " " ) ;
   /*----------------------------------------------------------------------*/
   wxRegEx re( R"=(\mclass="zone-dedicated-availability")="
               R"=(.*?\mdata-ref="([^"]+)")=" // Reference
               R"=(.*?\m(KS-\d*)<)="          // Name
               R"=(.*?<td.*?>(.*?)</td>)="    // CPU Brand & Type
               R"=(.*?<td.*?>(.*?)</td>)="    // CPU Cores/Threads
               R"=(.*?<td.*?>(.*?)</td>)="    // CPU Speed
               R"=(.*?<td.*?>(.*?)</td>)="    // Memory
               R"=(.*?<td.*?>(.*?)</td>)="    // Storage
               R"=(.*?<td.*?>(.*?)</td>)="    // Net speed
               R"=(.*?<td.*?>(.*?)</td>)="    // Net mask
               R"=(.*?<td.*?>(.*?)</td>)="    // Price
               R"=(.*?\mdata-dc="([^"]+)")=", // Data center
               wxRE_ADVANCED
             ) ;
   wxRegEx re_clean( R"=(<span[^>]*>|</span>|<br>)=", wxRE_ADVANCED ) ;
   wxRegEx re_clean_space( R"=(\s+)=", wxRE_ADVANCED ) ;
   /*----------------------------------------------------------------------*/
   const wxChar *p_c_data = s_data.c_str() ;
   size_t sz_start ;
   size_t sz_len   ;
   CServer server  ;
   int i_nb_server_selected = 0 ;
   for( ; re.Matches( p_c_data ) ; p_c_data += sz_start + sz_len )
   {
      /*--( Keep the found element to skip it later )----------------------*/
      if( !re.GetMatch( &sz_start, &sz_len, 0 ) )
      {  return -2 ; }
      /*-------------------------------------------------------------------*/
      CServer server ;
      server.m_s_ref  = re.GetMatch( p_c_data, 1 ) ;
      server.m_s_name = re.GetMatch( p_c_data, 2 ) ;
      /*--( Extract the different parts of the server description )--------*/
      const size_t co_sz_nb_match = re.GetMatchCount() ;
      wxString s_val ;
      for( size_t sz_num = 3 ; sz_num < co_sz_nb_match ; ++sz_num )
      {
         /*--( Extracting and "cleaning" a value )-------------------------*/
         s_val = re.GetMatch( p_c_data, sz_num ) ;
         re_clean_space.ReplaceAll( &s_val, " " ) ;
         re_clean.ReplaceAll( &s_val, "" ) ;
         s_val = s_val.Trim().Trim( false ) ;
         /*--( Price & DC are the last data. They are stored separately )--*/
         if( sz_num == co_sz_nb_match - 2 )
         {  server.m_s_price = s_val ; }
         else
         if( sz_num == co_sz_nb_match - 1 )
         {  server.m_s_dc = s_val ; }
         else
         {  server.m_s_spec += s_val ;
            if( sz_num < co_sz_nb_match - 3 ) { server.m_s_spec += " - " ; }
         }
         /*----------------------------------------------------------------*/
      }
      /*--( Create the vector element )------------------------------------*/
      m_vec_server.emplace_back( server ) ;
      auto it = m_vec_server.end() - 1 ;

      /*--( Restore previously entered data )------------------------------*/
      auto it_save = find_if( vec_server_save.begin(), vec_server_save.end(),
                              [ &server ]( const CServer &server_save )
                              {  return server.m_s_ref
                                     == server_save.m_s_ref ;
                              }
                            ) ;
      if( it_save == vec_server_save.end() )
      {  it->m_avail = CServer::AVA_UNKNOWN ; }
      else
      {  /*----------------------------------------------------------------*/
         it->m_boo_select = it_save->m_boo_select ;
         it->m_avail = it_save->m_avail ;
         if( it->m_boo_select ) { ++i_nb_server_selected ; }
         /*----------------------------------------------------------------*/
      }
      /*-------------------------------------------------------------------*/
   }

   /*----------------------------------------------------------------------*/
   SetItemCount( m_vec_server.size() ) ;
   refresh_all_items() ;
   /*----------------------------------------------------------------------*/
   wxString s_msg ;
   s_msg.Printf( '\t' + wxPLURAL( _( "%zu server found" ),
                                  _( "%zu servers found" ),
                                  m_vec_server.size()
                                 ),
                 m_vec_server.size()
               ) ;
   frame->log_message( s_msg ) ;

   /*-----------------------------------------------------------------------+
   ! If some servers are selected an availability refresh seems user        !
   ! friendly                                                               !
   +-----------------------------------------------------------------------*/
   if( i_nb_server_selected > 0 ) { launch_check_avail() ; }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::launch_check_avail()
{
   /*----------------------------------------------------------------------*/
   const wxString co_s_title
       = wxString::Format( _( "Refresh availabilities #%d" ),
                           ++m_i_nb_check_avail
                         ) ;
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   refresh_all_items() ;
   m_boo_check_batch_alert_done = false ;
   /*----------------------------------------------------------------------*/
   const wxString co_s_url
       = "https://www.ovh.com/engine/api/dedicated/server/"
         "availabilities?country=fr&hardware=" ;
   /*----------------------------------------------------------------------*/
   int i_nb_dl = 0 ;
   /*----------------------------------------------------------------------*/
   for( auto &server : m_vec_server )
   {
      /*-------------------------------------------------------------------*/
      if( !server.m_boo_select ) { continue ; }
      /*-------------------------------------------------------------------*/
      sc::CDownload dl( this, co_s_url + server.m_s_ref ) ;
      dl.get_dlbase().set_user_data( server.m_s_ref ) ;
      dl.get_dlbase().set_download_id( DID_AVAIL ) ;
      dl.thread_start() ;
      ++i_nb_dl ;
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   if( i_nb_dl == 0 )
   {  frame->log_error( wxString::Format( _( "%s: no server selected" ),
                                          co_s_title
                                        )
                      ) ;
   }
   else
   {  frame->log_message( wxString::Format( wxPLURAL( _( "%s: %d server" ),
                                                      _( "%s: %d servers" ),
                                                      i_nb_dl
                                                    ),
                                            co_s_title, i_nb_dl
                                          )
                        ) ;
   }
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*--( This function is supposed to be called once per server availability )*/
int CServerList::check_avail( int i_err_code,
                              const wxString &co_s_ref,
                              const wxString &co_s_data
                            )
{
   /*----------------------------------------------------------------------*/
   wxString co_s_title = _( "Availability" ) ;
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   if( i_err_code != sc::CDownloadBase::DE_NO_ERROR_DONE )
   {  frame->log_error( sc::CDownloadBase::st_get_err_msg( i_err_code ) ) ;
      return -1 ;
   }
   /*----------------------------------------------------------------------*/
   if( co_s_ref.empty() )
   {  frame->log_message( wxString::Format( _( "%s: no server reference !" ),
                                            co_s_title
                                          )
                        ) ;
      return -2 ;
   }
   /*----------------------------------------------------------------------*/
   if( co_s_data.empty() )
   {  frame->log_message( wxString::Format( _( "%s: no data" ), co_s_title
                                          )
                        ) ;
      return -3 ;
   }
   /*--( We have to find the server )--------------------------------------*/
   auto it = find_if( m_vec_server.begin(), m_vec_server.end(),
                      [ &co_s_ref ]( const CServer &server )
                      {  return server.m_s_ref == co_s_ref ; }
                     ) ;
   if( it == m_vec_server.end() )
   {  frame->log_message( wxString::Format( "\t" + _( "%s: %s unknown !" ),
                                            co_s_title, co_s_ref
                                          )
                        ) ;
      return -4 ;
   }

   /*--( Is this server available ? )--------------------------------------*/
   if( sc::match( co_s_data, R"=("availability":"\d)=" ) <= 0 )
   {  /*-------------------------------------------------------------------*/
      it->m_avail = CServer::AVA_KO ;
      /*-------------------------------------------------------------------*/
      frame->log_message( wxString::Format( "\t" + _( "%s NOT available" ),
                                            co_s_ref
                                          )
                        ) ;
      /*-------------------------------------------------------------------*/
   }
   else
   {
      /*--( Signal this availability )-------------------------------------*/
      it->m_avail = CServer::AVA_OK ;
      /*-------------------------------------------------------------------*/
      frame->log_message_ok( wxString::Format( "\t" + _( "%s AVAILABLE" ),
                                               co_s_ref
                                             )
                           ) ;
      /*--( Alert done only once per batch )-------------------------------*/
      if( !m_boo_check_batch_alert_done )
      {  wxGetApp().m_frame->alert_selected_server_avail() ;
         m_boo_check_batch_alert_done = true ;
      }
      /*-------------------------------------------------------------------*/
   }
   /*----------------------------------------------------------------------*/
   RefreshItem( it - m_vec_server.begin() ) ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::start_auto_check_avail()
{
   /*----------------------------------------------------------------------*/
   if( get_auto_check_avail_running() ) { return -1 ; }
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   const int co_i_sec  = wxGetApp().M_i_check_avail_period.get() ;
   m_timer_check_avail.Start( co_i_sec * 1000 ) ;
   frame->log_message( wxString::Format(
                    _( "Availability: automatic check done every %ds (%s)" ),
                    co_i_sec, sc::sec2hms( co_i_sec )
                                       )
                     ) ;
   /*----------------------------------------------------------------------*/
   frame->gui_init_auto_check_avail() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
int CServerList::stop_auto_check_avail()
{
   /*----------------------------------------------------------------------*/
   CFrame *frame = wxGetApp().m_frame ;
   /*----------------------------------------------------------------------*/
   m_timer_check_avail.Stop() ;
   frame->log_message( _( "Availability: automatic check stopped" ) ) ;
   /*----------------------------------------------------------------------*/
   frame->gui_init_auto_check_avail() ;
   /*----------------------------------------------------------------------*/
   return 0 ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
wxString CServerList::get_url( long l_item )
{
   /*----------------------------------------------------------------------*/
   const CServer *p_server = get_server( l_item ) ;
   if( p_server == nullptr ) { return wxEmptyString ; }
   /*----------------------------------------------------------------------*/
   const str_region *p_region = get_region() ;
   if( p_region == nullptr ) { return wxEmptyString ; }
   /*----------------------------------------------------------------------*/
   return wxString::Format( "%s/%s%s", m_co_p_c_main_url,
                            p_region->m_p_c_url_order,
                            p_server->m_s_ref
                          ) ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
const str_region *CServerList::get_region( const wxString &s_region ) const
{
   /*----------------------------------------------------------------------*/
   if( !s_region.empty() )
   {  auto it = std::find_if( m_vec_region.begin(), m_vec_region.end(),
                              [ &s_region ]( const str_region &region )
                              {  return region.m_s_code == s_region ; }
                            ) ;
      if( it != m_vec_region.end() ) { return &*it ; }
   }
   /*----------------------------------------------------------------------*/
   return nullptr ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/
const str_region *CServerList::get_region() const
{
   /*----------------------------------------------------------------------*/
   wxString s_region = wxGetApp().M_s_region.get() ;
   const str_region *p_region ;

   /*----------------------------------------------------------------------*/
   if( ( p_region = get_region( s_region ) ) != nullptr )
   {  return p_region ; }

   /*--( Default to current system language / region )---------------------*/
   s_region = wxGetApp().get_canonical_name().Before( '_' ) ;
   if( ( p_region = get_region( s_region ) ) != nullptr )
   {  return p_region ; }

   /*----------------------------------------------------------------------*/
   return nullptr ;
   /*----------------------------------------------------------------------*/
}

/*-------------------------------------------------------------------------*/



/*==========================================================================+
!                       End of file CServerList.cpp                         !
+==========================================================================*/
